#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import json

import jwt
import requests


class Credentials:
    PropertyIdToken = 'id_token'
    """
    Credentials represents an Fabric Identity token used for authenticating/authorization on behalf of the experimentor
    """
    def __init__(self, *, id_token: str = None, properties: dict = None):
        from fabric.actor.core.container.globals import GlobalsSingleton
        self.id_token = id_token
        self.jwks_url = GlobalsSingleton.get().get_config().get("oauth", "oauth-jwks-url")

        if properties is not None and self.PropertyIdToken in properties:
            self.id_token = properties[self.PropertyIdToken]

    def validate(self):
        try:
            response = requests.get(self.jwks_url)
            if response.status_code != 200:
                return
            jwks = response.json()
            public_keys = {}
            for jwk in jwks['keys']:
                kid = jwk['kid']
                public_keys[kid] = jwt.algorithms.RSAAlgorithm.from_jwk(json.dumps(jwk))

            kid = jwt.get_unverified_header(self.id_token)['kid']
            key = public_keys[kid]
            options = {'verify_aud': False}
            payload = jwt.decode(self.id_token, key=key, algorithms=['RS256'], options=options)
            return payload
        except Exception as e:
            raise e



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import traceback
from typing import TYPE_CHECKING

from fabric.actor.boot.inventory.pool_creator import PoolCreator
from fabric.actor.core.apis.i_authority import IAuthority
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.resource_pool_attribute_descriptor import ResourcePoolAttributeDescriptor, \
    ResourcePoolAttributeType
from fabric.actor.core.common.resource_pool_descriptor import ResourcePoolDescriptor
from fabric.actor.core.container.remote_actor_cache import RemoteActorCacheSingleton, RemoteActorCache
from fabric.actor.core.core.authority import Authority
from fabric.actor.core.core.broker import Broker
from fabric.actor.core.core.controller import Controller
from fabric.actor.core.delegation.simple_resource_ticket_factory import SimpleResourceTicketFactory
from fabric.actor.core.kernel.slice_factory import SliceFactory
from fabric.actor.core.manage.management_utils import ManagementUtils
from fabric.actor.core.manage.messages.client_mng import ClientMng
from fabric.actor.core.plugins.base_plugin import BasePlugin
from fabric.actor.core.plugins.config.config import Config
from fabric.actor.core.plugins.db.server_actor_database import ServerActorDatabase
from fabric.actor.core.plugins.substrate.authority_substrate import AuthoritySubstrate
from fabric.actor.core.plugins.substrate.substrate import Substrate
from fabric.actor.core.plugins.substrate.db.substrate_actor_database import SubstrateActorDatabase
from fabric.actor.core.policy.authority_calendar_policy import AuthorityCalendarPolicy
from fabric.actor.core.policy.broker_simpler_units_policy import BrokerSimplerUnitsPolicy
from fabric.actor.core.policy.controller_simple_policy import ControllerSimplePolicy
from fabric.actor.core.proxies.kafka.translate import Translate
from fabric.actor.core.util.id import ID
from fabric.actor.core.util.reflection_utils import ReflectionUtils
from fabric.actor.core.util.resource_type import ResourceType
from fabric.actor.core.core.actor import Actor
from fabric.actor.security.auth_token import AuthToken
from fabric.message_bus.messages.claim_resources_avro import ClaimResourcesAvro
from fabric.actor.core.apis.i_actor import ActorType

if TYPE_CHECKING:
    from fabric.actor.boot.configuration import Configuration, ActorConfig, PolicyConfig, Peer
    from fabric.actor.core.apis.i_actor import IActor
    from fabric.actor.core.apis.i_mgmt_actor import IMgmtActor


class ExportAdvertisement:
    def __init__(self, *, exporter: IMgmtActor, client: ClientMng, delegation: str, topic: str):
        self.exporter = exporter
        self.client = client
        self.delegation = delegation
        self.client_topic = topic

        self.exported = None


class ExportInfo:
    def __init__(self, *, exporter: IMgmtActor, client: ClientMng, units: int, rtype: ResourceType, topic: str):
        self.exporter = exporter
        self.client = client
        self.units = units
        self.rtype = rtype
        self.client_topic = topic

        self.exported = None
        self.start = None
        self.end = None


class ConfigurationProcessor:
    def __init__(self, *, config: Configuration):
        from fabric.actor.core.container.globals import GlobalsSingleton
        self.logger = GlobalsSingleton.get().get_logger()
        self.config = config
        self.actor = None
        self.to_export = []
        self.to_advertise = []
        self.pools = {}
        self.resources = None
        self.aggregate_delegation_models = None

    def process(self):
        try:
            self.create_actor()
            self.initialize_actor()
            self.logger.info("There are {} actors".format(Actor.actor_count))
            self.register_actor()
            self.create_default_slice()
            self.populate_inventory_neo4j()
            self.populate_inventory()
            self.recover_actor()
            self.enable_ticking()
            self.process_topology()
            self.logger.info("Processing exports with actor count {}".format(Actor.actor_count))
            self.process_exports()
            self.process_advertise()
            self.logger.info("Processing exports completed")
            self.logger.info("Processing claims with actor count {}".format(Actor.actor_count))
            self.process_claims()
            self.logger.info("Processing claims completed")
        except Exception as e:
            self.logger.error(traceback.format_exc())
            raise Exception("Unexpected error while processing configuration {}".format(e))
        self.logger.info("Finished instantiating actors.")
        print("End process")

    def create_actor(self):
        try:
            if self.config.get_actor() is not None:
                self.logger.debug("Creating Actor: name={}".format(self.config.get_actor().get_name()))
                self.actor = self.do_common(actor_config=self.config.get_actor())
                self.do_specific(actor=self.actor, config=self.config.get_actor())
        except Exception as e:
            raise Exception("Unexpected error while creating actor {}".format(e))

    def do_common(self, *, actor_config: ActorConfig):
        actor = self.make_actor_instance(actor_config=actor_config)
        actor.set_plugin(plugin=self.make_plugin_instance(actor=actor, actor_config=actor_config))
        actor.set_policy(policy=self.make_actor_policy(actor=actor, config=actor_config))
        return actor

    def make_actor_instance(self, *, actor_config: ActorConfig) -> IActor:
        actor_type = ActorType.get_actor_type_from_string(actor_type=actor_config.get_type())
        actor = None
        if actor_type == ActorType.Orchestrator:
            actor = Controller()
        elif actor_type == ActorType.Broker:
            actor = Broker()
        elif actor_type == ActorType.Authority:
            actor = Authority()
        else:
            raise Exception("Unsupported actor type: {}".format(actor_type))

        if actor is not None:
            actor_guid = ID()
            if actor_config.get_guid() is not None:
                actor_guid = ID(id=actor_config.get_guid())
            auth_token = AuthToken(name=actor_config.get_name(), guid=actor_guid)
            actor.set_identity(token=auth_token)
            if actor_config.get_description() is not None:
                actor.set_description(description=actor_config.get_description())

            from fabric.actor.core.container.globals import GlobalsSingleton
            actor.set_actor_clock(clock=GlobalsSingleton.get().get_container().get_actor_clock())

        return actor

    def make_plugin_instance(self, *, actor: IActor, actor_config: ActorConfig):
        plugin = None
        if actor.get_plugin() is None:
            if actor.get_type() == ActorType.Authority:
                # TODO replacement of ANT CONFIG
                plugin = AuthoritySubstrate(actor=actor, db=None, config=Config())

            elif actor.get_type() == ActorType.Orchestrator:
                # TODO replacement of ANT CONFIG
                plugin = Substrate(actor=actor, db=None, config=Config())

            elif actor.get_type() == ActorType.Broker:
                plugin = BasePlugin(actor=actor, db=None, config=Config())

        if plugin is None:
            raise Exception("Cannot instantiate shirako plugin for actor: {}".format(actor_config.get_name()))

        if plugin.get_database() is None:
            db = None
            user = self.config.get_global_config().get_database()[Constants.PropertyConfDbUser]
            password = self.config.get_global_config().get_database()[Constants.PropertyConfDbPassword]
            db_host = self.config.get_global_config().get_database()[Constants.PropertyConfDbHost]
            db_name = self.config.get_global_config().get_database()[Constants.PropertyConfDbName]
            if isinstance(plugin, Substrate):
                db = SubstrateActorDatabase(user=user, password=password, database=db_name, db_host=db_host, logger=self.logger)
            else:
                db = ServerActorDatabase(user=user, password=password, database=db_name, db_host=db_host, logger=self.logger)

            plugin.set_database(db=db)

        ticket_factory = SimpleResourceTicketFactory()
        ticket_factory.set_actor(actor=actor)
        plugin.set_ticket_factory(ticket_factory=ticket_factory)
        return plugin

    def make_actor_policy(self, *, actor: IActor, config: ActorConfig):
        policy = None
        if config.get_policy() is None:
            if actor.get_type() == ActorType.Authority:
                policy = self.make_site_policy(config=config)
            elif actor.get_type() == ActorType.Broker:
                policy = BrokerSimplerUnitsPolicy()
            elif actor.get_type() == ActorType.Orchestrator:
                policy = ControllerSimplePolicy()
        else:
            policy = self.make_policy(policy=config.get_policy())

        if policy is None:
            raise Exception("Could not instantiate policy for actor: {}".format(config.get_name()))
        return policy

    def make_site_policy(self, *, config: ActorConfig):
        # TODO KOMAL
        #if config.get_controls() is None or len(config.get_controls()) == 0:
        #    raise Exception("Missing authority policy but no control has been specified")

        policy = AuthorityCalendarPolicy()
        for c in config.get_controls():
            try:
                if c.get_module_name() is None or c.get_class_name() is None:
                    raise Exception("Missing control class name")

                control = ReflectionUtils.create_instance(module_name=c.get_module_name(),
                                                          class_name=c.get_class_name())

                if c.get_type() is None:
                    raise Exception("No type specified for control")

                control.add_type(rtype=ResourceType(resource_type=c.get_type()))
                policy.register_control(control=control)
            except Exception as e:
                traceback.print_exc()
                raise Exception("Could not create control {}".format(e))
        return policy

    def make_policy(self, *, policy: PolicyConfig):
        if policy.get_class_name() is None or policy.get_module_name() is None:
            raise Exception("Policy is missing class name")

        return ReflectionUtils.create_instance(module_name=policy.get_module_name(), class_name=policy.get_class_name())

    def do_specific(self, *, actor: IActor, config: ActorConfig):
        if isinstance(actor, IAuthority):
            self.pools = self.read_resource_pools(config=config)
            self.resources = self.read_resource_pools2(config=config)

    def read_resource_pools(self, *, config: ActorConfig) -> dict:
        result = {}
        pools = config.get_pools()
        if pools is None or len(pools) == 0:
            return result

        for p in pools:
            descriptor = ResourcePoolDescriptor()
            descriptor.set_resource_type(rtype=ResourceType(resource_type=p.get_type()))
            descriptor.set_resource_type_label(rtype_label=p.get_label())
            descriptor.set_units(units=p.get_units())
            descriptor.set_start(start=p.get_start())
            descriptor.set_end(end=p.get_end())
            descriptor.set_pool_factory_class(factory_class=p.get_factory_class())
            descriptor.set_pool_factory_module(factory_module=p.get_factory_module())
            handler = p.get_handler()
            if handler is not None:
                descriptor.set_handler_class(handler_class=handler.get_class_name())
                descriptor.set_handler_module(module=handler.get_module_name())
                descriptor.set_handler_properties(properties=handler.get_properties())

            descriptor.pool_properties = p.get_properties()

            for attr in p.get_attributes():
                attribute = ResourcePoolAttributeDescriptor()
                attribute.set_key(value=attr.get_key())
                attribute.set_value(value=attr.get_value())
                if attr.get_type().lower() == "integer":
                    attribute.set_type(rtype=ResourcePoolAttributeType.INTEGER)
                elif attr.get_type().lower() == "string":
                    attribute.set_type(rtype=ResourcePoolAttributeType.STRING)
                elif attr.get_type().lower() == "neo4j":
                    attribute.set_type(rtype=ResourcePoolAttributeType.NEO4J)
                elif attr.get_type().lower() == "class":
                    attribute.set_type(rtype=ResourcePoolAttributeType.CLASS)
                else:
                    raise Exception("Unsupported attribute type: {}".format(attr.get_type()))
                descriptor.add_attribute(attribute=attribute)

            result[descriptor.get_resource_type()] = descriptor
        return result

    def read_resource_pools2(self, *, config: ActorConfig) -> dict:
        result = {}
        resources = config.get_resources()
        if resources is None or len(resources) == 0:
            return result

        for r in resources:
            descriptor = ResourcePoolDescriptor()
            descriptor.set_resource_type(rtype=ResourceType(resource_type=r.get_type()))
            descriptor.set_resource_type_label(rtype_label=r.get_label())

            # TODO for now it holds the control name
            descriptor.set_pool_factory_class(factory_class=r.get_control().get_class_name())
            descriptor.set_pool_factory_module(factory_module=r.get_control().get_module_name())

            handler = r.get_handler()
            if handler is not None:
                descriptor.set_handler_class(handler_class=handler.get_class_name())
                descriptor.set_handler_module(module=handler.get_module_name())
                descriptor.set_handler_properties(properties=handler.get_properties())

            descriptor.pool_properties = r.get_properties()

            for attr in r.get_attributes():
                attribute = ResourcePoolAttributeDescriptor()
                attribute.set_key(value=attr.get_key())
                attribute.set_value(value=attr.get_value())
                if attr.get_type().lower() == "integer":
                    attribute.set_type(rtype=ResourcePoolAttributeType.INTEGER)
                elif attr.get_type().lower() == "string":
                    attribute.set_type(rtype=ResourcePoolAttributeType.STRING)
                elif attr.get_type().lower() == "neo4j":
                    attribute.set_type(rtype=ResourcePoolAttributeType.NEO4J)
                elif attr.get_type().lower() == "class":
                    attribute.set_type(rtype=ResourcePoolAttributeType.CLASS)
                else:
                    raise Exception("Unsupported attribute type: {}".format(attr.get_type()))
                descriptor.add_attribute(attribute=attribute)

            result[descriptor.get_resource_type()] = descriptor
        return result

    def initialize_actor(self):
        try:
            Actor.actor_count += 1
            self.actor.initialize()
        except Exception as e:
            raise Exception("Actor failed to initialize: {} {}".format(self.actor.get_name(), e))

    def register_actor(self):
        try:
            from fabric.actor.core.container.globals import GlobalsSingleton
            GlobalsSingleton.get().get_container().register_actor(actor=self.actor)
        except Exception as e:
            raise Exception("Could not register actor: {} {}".format(self.actor.get_name(), e))

    def create_default_slice(self):
        if self.actor.get_type() != ActorType.Authority:
            slice_obj = SliceFactory.create(slice_id=ID(), name=self.actor.get_name())
            slice_obj.set_inventory(value=True)
            try:
                self.actor.register_slice(slice_object=slice_obj)
            except Exception as e:
                Exception("Could not create default slice for actor: {} {}".format(self.actor.get_name(), e))

    def populate_inventory(self):
        if isinstance(self.actor, IAuthority):
            if isinstance(self.actor.get_plugin(), AuthoritySubstrate):
                creator = PoolCreator(substrate=self.actor.get_plugin(), pools=self.pools,
                                      neo4j_config=self.config.get_global_config().get_neo4j_config())
                creator.process()

    def populate_inventory_neo4j(self):
        if isinstance(self.actor, IAuthority):
            if isinstance(self.actor.get_plugin(), AuthoritySubstrate):
                creator = PoolCreator(substrate=self.actor.get_plugin(), pools=self.resources,
                                      neo4j_config=self.config.get_global_config().get_neo4j_config())
                self.aggregate_delegation_models = creator.process_neo4j(actor_name=self.actor.get_name(),
                                                                         substrate_file=self.config.get_actor().
                                                                         get_substrate_file())

    def recover_actor(self):
        try:
            self.actor.recover()
        except Exception as e:
            raise Exception("Recovery failed for actor: {}".format(self.actor.get_name()))

    def enable_ticking(self):
        from fabric.actor.core.container.globals import GlobalsSingleton
        GlobalsSingleton.get().get_container().register(tickable=self.actor)

    def process_topology(self):
        if self.config.get_peers() is None or len(self.config.get_peers()) == 0:
            self.logger.debug("No peers specified")
        else:
            self.create_proxies(peers=self.config.get_peers())

    def process_exports(self):
        for ei in self.to_export:
            self.export(info=ei)

    def process_advertise(self):
        if self.aggregate_delegation_models is None and len(self.to_advertise) > 0:
            raise Exception("No delegations found in Aggregate Resource Model")

        # TODO re-enable check after testing
        #if len(self.aggregate_delegation_models) != len(self.to_advertise):
        #    raise Exception("Number of delegations in Aggregate Resource Model does not match Peers with delegations")

        for ei in self.to_advertise:
            self.advertise(info=ei)

    def process_claims(self):
        self.logger.debug("process_claims {}".format(len(self.to_export)))
        for ei in self.to_export:
            self.claim(info=ei)

    def create_proxies(self, *, peers: list):
        for e in peers:
            self.process_peer(peer=e)

    def vertex_to_registry_cache(self, *, peer: Peer):
        self.logger.debug("Adding vertex for {}".format(peer.get_name()))

        if peer.get_name() is None:
            raise Exception("Actor must specify a name")

        if peer.get_guid() is None:
            raise Exception("Actor must specify a guid")

        protocol = Constants.ProtocolLocal
        kafka_topic = None
        if peer.get_kafka_topic() is not None:
            protocol = Constants.ProtocolKafka
            kafka_topic = peer.get_kafka_topic()

        actor_type = peer.get_type().lower()

        entry = {
            RemoteActorCache.ActorName: peer.get_name(),
            RemoteActorCache.ActorGuid: ID(id=peer.get_guid()),
            RemoteActorCache.ActorType: actor_type,
            RemoteActorCache.ActorProtocol: protocol
        }
        if kafka_topic is not None:
            entry[RemoteActorCache.ActorLocation] = kafka_topic

        RemoteActorCacheSingleton.get().add_partial_cache_entry(guid=ID(id=peer.get_guid()), entry=entry)

    def process_peer(self, *, peer: Peer):
        from_guid = ID(id=peer.get_guid())
        from_type = ActorType.get_actor_type_from_string(actor_type=peer.get_type())
        to_guid = self.actor.get_guid()
        to_type = self.actor.get_type()

        # We only like peers broker->site and service->broker
        # Reverse the peer if it connects site->broker or broker->service

        if from_type == ActorType.Authority and to_type == ActorType.Broker:
            temp = from_guid
            from_guid = to_guid
            to_guid = temp
            temp = from_type
            from_type = to_type
            to_type = temp

        if from_type == ActorType.Broker and to_type == ActorType.Orchestrator:
            temp = from_guid
            from_guid = to_guid
            to_guid = temp
            temp = from_type
            from_type = to_type
            to_type = temp

        # peers between actors of same type aren't allowed unless the actors are both brokers
        if from_type == to_type and from_type != ActorType.Broker:
            raise Exception("Invalid peer type: broker can only talk to broker, orchestrator or site authority")

        container = ManagementUtils.connect(caller=self.actor.get_identity())
        to_mgmt_actor = container.get_actor(guid=to_guid)
        if to_mgmt_actor is None:
            self.logger.debug("to_mgmt_actor={} to_guid={}".format(type(to_mgmt_actor), to_guid))
            if container.get_last_error() is not None:
                self.logger.error(container.get_last_error())
        from_mgmt_actor = container.get_actor(guid=from_guid)
        if from_mgmt_actor is None:
            self.logger.debug("from_mgmt_actor={} from_guid={}".format(type(from_mgmt_actor), from_guid))
            if container.get_last_error() is not None:
                self.logger.error(container.get_last_error())

        self.vertex_to_registry_cache(peer=peer)

        try:
            client = RemoteActorCacheSingleton.get().establish_peer(from_guid=from_guid, from_mgmt_actor=from_mgmt_actor,
                                                                    to_guid=to_guid, to_mgmt_actor=to_mgmt_actor)
            self.logger.debug("Client returned {}".format(client))
            if client is not None:
                self.parse_exports(peer=peer, client=client, mgmt_actor=to_mgmt_actor)
        except Exception as e:
            Exception("Could not process exports from: {} to {}. e= {}" .format(peer.get_guid(), self.actor.get_guid(), e))

    def parse_exports(self, *, peer: Peer, client: ClientMng, mgmt_actor: IMgmtActor):
        for rset in peer.get_rsets():
            info = ExportInfo(exporter=mgmt_actor, client=client, units=rset.get_units(),
                              rtype=ResourceType(resource_type=rset.get_type()), topic=peer.get_kafka_topic())
            if rset.get_start() is not None:
                info.start = rset.get_start()
            if rset.get_end() is not None:
                info.end = rset.get_end()

            self.to_export.append(info)

        if peer.get_delegation() is not None:
            info = ExportAdvertisement(exporter=mgmt_actor, client=client,
                                       delegation=peer.get_delegation(),
                                       topic=peer.get_kafka_topic())
            self.to_advertise.append(info)

    def export(self, *, info: ExportInfo):
        from fabric.actor.core.container.globals import GlobalsSingleton
        now = GlobalsSingleton.get().get_container().get_current_cycle()
        start = info.start
        if start is None:
            start = GlobalsSingleton.get().get_container().cycle_start_date(cycle=now)
        end = info.end
        if end is None:
            # export for one year
            length = 1000 * 60 * 60 * 24 * 365
            end = GlobalsSingleton.get().get_container().cycle_end_date(cycle=now + length)

        self.logger.debug("Using Server Actor {} to export resources".format(info.exporter.__class__.__name__))

        info.exported = info.exporter.export_resources(rtype=info.rtype, start=start,
                                                       end=end, units=info.units, ticket_properties=None,
                                                       resource_properties=None, source_ticket_id=None,
                                                       client=AuthToken(name=info.client.get_name(),
                                                                        guid=ID(id=info.client.get_guid())))

        if info.exported is None:
            raise Exception("Could not export resources from actor: {} to actor: {} Error = {}".
                            format(info.exporter.get_name(), info.client.get_name(), info.exporter.get_last_error()))

    def advertise(self, *, info: ExportAdvertisement):
        self.logger.debug("Using Server Actor {} to export resources".format(info.exporter.__class__.__name__))

        delegation = self.aggregate_delegation_models.get(info.delegation, None)

        if delegation is None:
            raise Exception("Aggregate Delegation Model not found for delegation: {}".format(info.delegation))

        info.exported = info.exporter.advertise_resources(delegation=delegation,
                                                          client=AuthToken(name=info.client.get_name(),
                                                                        guid=ID(id=info.client.get_guid())))

        if info.exported is None:
            raise Exception("Could not export resources from actor: {} to actor: {} Error = {}".
                            format(info.exporter.get_name(), info.client.get_name(), info.exporter.get_last_error()))

    # TODO needs to be fixed
    def claim(self, *, info: ExportInfo):
        if info.exported is None:
            self.logger.error("No reservation to export from {} to {}".format(info.exporter.get_name(), info.client.get_name()))
            return

        self.logger.debug("Claiming resources from {} to {}".format(info.exporter.get_name(), info.client.get_name()))

        container = ManagementUtils.connect(caller=self.actor.get_identity())
        client_mgmt_actor = container.get_actor(guid=ID(id=info.client.get_guid()))

        if client_mgmt_actor is None:
            self.logger.info("{} is a remote client. Not performing claim".format(info.client.get_name()))
            #self.trigger_remote_claim(info)
            return

        self.logger.info("Claiming resources from {} to {}".format(info.exporter.get_name(), info.client.get_name()))

        reservation = client_mgmt_actor.claim_resources(broker=info.exporter.get_guid(), rid=info.exported)

        if reservation is not None:
            self.logger.info("Successfully initiated claim for resources from {} to {}".format(info.exporter.get_name(),
                                                                                               info.client.get_name()))
        else:
            self.logger.error("Could not initiate claim for resources from {} to {}".format(info.exporter.get_name(),
                                                                                            info.client.get_name()))

    def trigger_remote_claim(self, *, info: ExportInfo):
        try:
            claim_req = ClaimResourcesAvro()
            claim_req.guid = info.client.guid
            claim_req.auth = Translate.translate_auth_to_avro(auth=self.actor.get_identity())
            claim_req.broker_id = str(info.exporter.get_guid())
            claim_req.reservation_id = str(info.exported)
            claim_req.message_id = "test_claim_1"
            claim_req.callback_topic = self.config.actor.get_kafka_topic()
            claim_req.slice_id = "null"

            # create a producer
            from fabric.actor.core.container.globals import GlobalsSingleton
            producer = GlobalsSingleton.get().get_kafka_producer()
            if producer.produce_sync(topic=info.client_topic, record=claim_req):
                self.logger.debug("Message {} written to {}".format(claim_req.name, info.client_topic))
            else:
                self.logger.error("Failed to send message {} to {}".format(claim_req.name, info.client_topic))
        except Exception as e:
            traceback.print_exc()
            self.logger.error(e)
            print(e)


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import traceback
from typing import TYPE_CHECKING

import threading

from fabric.message_bus.consumer import AvroConsumerApi
from fabric.message_bus.messages.message import IMessageAvro

if TYPE_CHECKING:
    from fabric.actor.core.proxies.kafka.services.actor_service import ActorService
    from fabric.actor.core.manage.kafka.services.kafka_actor_service import KafkaActorService


class MessageService(AvroConsumerApi):
    def __init__(self, *, kafka_service: ActorService, kafka_mgmt_service: KafkaActorService, conf: dict, key_schema,
                 record_schema, topics, batchSize=5, logger=None):
        super().__init__(conf=conf, key_schema=key_schema, record_schema=record_schema, topics=topics,
                         batchSize=batchSize, logger=logger)
        self.thread_lock = threading.Lock()
        self.thread = None
        self.kafka_service = kafka_service
        self.kafka_mgmt_service = kafka_mgmt_service

    def start(self):
        try:
            self.thread_lock.acquire()
            if self.thread is not None:
                raise Exception("This Message Service has already been started")

            self.thread = threading.Thread(target=self.consume_auto)
            self.thread.setName("MessageService")
            self.thread.setDaemon(True)
            self.thread.start()
            self.logger.debug("Message service has been started")
        finally:
            self.thread_lock.release()

    def stop(self):
        self.shutdown()
        try:
            self.thread_lock.acquire()
            temp = self.thread
            self.thread = None
            if temp is not None:
                self.logger.warning("It seems that the Message Service thread is running. Interrupting it")
                try:
                    temp.join()
                except Exception as e:
                    self.logger.error("Could not join actor thread {}".format(e))
                finally:
                    self.thread_lock.release()
        finally:
            if self.thread_lock is not None and self.thread_lock.locked():
                self.thread_lock.release()

    def handle_message(self, *, message: IMessageAvro):
        try:
            if message.get_message_name() == IMessageAvro.ClaimResources or \
                    message.get_message_name() == IMessageAvro.ReclaimResources or \
                    message.get_message_name() == IMessageAvro.GetSlicesRequest or \
                    message.get_message_name() == IMessageAvro.GetReservationsRequest or \
                    message.get_message_name() == IMessageAvro.GetReservationsStateRequest or \
                    message.get_message_name() == IMessageAvro.GetDelegations or \
                    message.get_message_name() == IMessageAvro.GetReservationUnitsRequest or \
                    message.get_message_name() == IMessageAvro.GetUnitRequest or \
                    message.get_message_name() == IMessageAvro.GetPoolInfoRequest or \
                    message.get_message_name() == IMessageAvro.AddSlice or \
                    message.get_message_name() == IMessageAvro.UpdateSlice or \
                    message.get_message_name() == IMessageAvro.RemoveSlice or \
                    message.get_message_name() == IMessageAvro.CloseReservations or \
                    message.get_message_name() == IMessageAvro.UpdateReservation or \
                    message.get_message_name() == IMessageAvro.RemoveReservation or \
                    message.get_message_name() == IMessageAvro.ExtendReservation:
                self.kafka_mgmt_service.process(message=message)
            else:
                self.kafka_service.process(message=message)
        except Exception as e:
            traceback.print_exc()
            self.logger.error(e)
            self.logger.error("Discarding the incoming message {}".format(message))





#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations
from typing import TYPE_CHECKING

from fabric.actor.core.apis.i_delegation import IDelegation
from fabric.actor.core.util.id import ID

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_actor import IActor
    from fabric.actor.core.apis.i_database import IDatabase
    from fabric.actor.core.apis.i_reservation import IReservation
    from fabric.actor.core.apis.i_slice import ISlice
    from fabric.actor.core.core.actor import Actor
    from fabric.actor.core.apis.i_resource_ticket_factory import IResourceTicketFactory
    from fabric.actor.core.plugins.config.config_token import ConfigToken
    from fabric.actor.core.util.resource_data import ResourceData
    from fabric.actor.security.auth_token import AuthToken

from fabric.actor.core.apis.i_actor import ActorType
from fabric.actor.core.apis.i_actor_event import IActorEvent
from fabric.actor.core.apis.i_base_plugin import IBasePlugin
from fabric.actor.core.delegation.simple_resource_ticket_factory import SimpleResourceTicketFactory
from fabric.actor.core.kernel.slice_factory import SliceFactory
from fabric.actor.core.plugins.config.config import Config


class BasePlugin(IBasePlugin):
    """
    The base implementation for actor-specific extensions.
    """
    PropertyConfig = "PluginConfig"
    PropertyConfigProperties = "PluginConfigProperties"
    PropertyActorName = "PluginActorName"

    def __init__(self, *, actor: Actor, db: IDatabase, config: Config):
        super().__init__()
        self.db = db
        self.config = config
        self.config_properties = None
        self.actor = actor
        self.logger = None
        self.from_config = False
        self.ticket_factory = None
        self.initialized = False

    def __getstate__(self):
        state = self.__dict__.copy()
        state['actor_id'] = self.actor.get_guid()
        del state['logger']
        del state['ticket_factory']
        del state['actor']
        del state['initialized']

        return state

    def __setstate__(self, state):
        actor_id = state['actor_id']
        # TODO fetch actor via actor_id
        del state['actor_id']
        self.__dict__.update(state)

    def initialize(self):
        if not self.initialized:
            try:
                if self.actor is None:
                    raise Exception("Missing actor")

                if self.ticket_factory is None:
                    self.ticket_factory = self.make_ticket_factory()

                if self.db is not None:
                    self.db.set_logger(logger=self.logger)
                    self.db.set_actor_name(name=self.actor.get_name())
                    # TODO
                    self.db.set_reset_state(state=True)
                    self.db.initialize()

                self.ticket_factory.initialize()
                self.initialized = True
            except Exception as e:
                raise e

    def configure(self, *, properties):
        self.config_properties = properties
        self.from_config = True

    def actor_added(self):
        if self.db is not None:
            self.db.actor_added()

        if self.config is not None:
            self.config.set_slices_plugin(plugin=self)
            self.config.initialize()

    def recovery_starting(self):
        return

    def restart_configuration_actions(self, *, reservation: IReservation):
        return

    def revisit(self, *, slice_obj: ISlice = None, reservation: IReservation = None, delegation: IDelegation = None):
        return

    def recovery_ended(self):
        return

    def create_slice(self, *, slice_id: ID, name: str, properties: ResourceData):
        slice_obj = SliceFactory.create(slice_id=slice_id, name=name, data=properties)
        return slice_obj

    def release_slice(self, *, slice_obj: ISlice):
        return

    def validate_incoming(self, *, reservation: IReservation, auth: AuthToken):
        return True

    def process_configuration_complete(self, *, token: ConfigToken, properties: dict):
        target = properties[Config.PropertyTargetName]
        unsupported = False

        if target == Config.TargetCreate:
            self.process_create_complete(token=token, properties=properties)
        elif target == Config.TargetDelete:
            self.process_delete_complete(token=token, properties=properties)
        elif target == Config.TargetModify:
            self.process_modify_complete(token=token, properties=properties)
        else:
            unsupported = True
            self.logger.warning("Unsupported target in configurationComplete(): {}".format(target))

        if not unsupported:
            self.actor.get_policy().configuration_complete(action=target, token=token, out_properties=properties)

    class ConfigurationCompleteEvent(IActorEvent):
        def __init__(self, *, token: ConfigToken, properties: dict, outer_class):
            self.token = token
            self.properties = properties
            self.outer_class = outer_class

        def process(self):
            self.outer_class.process_configuration_complete(token=self.token, properties=self.properties)

    def configuration_complete(self, *, token: ConfigToken, properties: dict):
        self.actor.queue_event(incoming=BasePlugin.ConfigurationCompleteEvent(token=token, properties=properties,
                                                                              outer_class=self))

    def get_actor(self):
        return self.actor

    def get_config(self):
        return self.config

    def get_database(self) -> IDatabase:
        return self.db

    def get_logger(self):
        return self.logger

    def make_ticket_factory(self) -> IResourceTicketFactory:
        ticket_factory = SimpleResourceTicketFactory()
        ticket_factory.set_actor(actor=self.actor)
        return ticket_factory

    def process_create_complete(self, *, token: ConfigToken, properties: dict):
        return

    def process_delete_complete(self, *, token: ConfigToken, properties: dict):
        return

    def process_modify_complete(self, *, token: ConfigToken, properties: dict):
        return

    def set_actor(self, *, actor: IActor):
        self.actor = actor

    def set_config(self, *, config: Config):
        self.config = config

    def set_database(self, *, db: IDatabase):
        self.db = db

    def set_logger(self, *, logger):
        self.logger = logger

    def get_ticket_factory(self) -> IResourceTicketFactory:
        return self.ticket_factory

    def set_ticket_factory(self, *, ticket_factory):
        self.ticket_factory = ticket_factory

    def get_config_properties(self) -> dict:
        return self.config_properties

    def is_site_authority(self):
        return self.actor.get_type() == ActorType.Authority




#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric.actor.core.util.id import ID
from fabric.actor.security.auth_token import AuthToken


class Guard:
    PropertyOwner = "owner"
    PropertyObjectId = "object_id"

    def __init__(self, *, properties: dict = None):
        self.owner = None
        self.object_id = None

        if properties is not None:
            if self.PropertyOwner in properties:
                self.owner = AuthToken(properties=properties[self.PropertyOwner])

            if self.PropertyObjectId in properties:
                self.object_id = ID(id=properties[self.PropertyObjectId])

    def check_reserve(self, *, requester: AuthToken):
        self.check_privelege(requester=requester)

    def check_update(self, *, requester:AuthToken):
        self.check_privelege(requester=requester)

    def check_privelege(self, *, requester:AuthToken):
        # TDB add PDP validatation
        return

    def check_owner(self, *, auth: AuthToken):
        if self.owner is not None:
            if self.owner != auth:
                raise Exception("Authorization Exception")

    def set_owner(self, *, owner:AuthToken):
        self.owner = owner

    def set_object_id(self, *, object_id: ID):
        self.object_id = object_id




#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric.actor.core.util.resource_type import ResourceType


class SiteResourceTypes:
    def __init__(self, *, domain: str):
        self.domain = domain
        self.map = {}

    def get_domain(self) -> str:
        return self.domain

    def get_resources(self):
        return self.map.values()

    def get_resource(self, *, rtype: ResourceType):
        return self.map.get(rtype, None)

    def get_default_resource(self):
        next(iter(self.map.values()))

    def add_resource(self, *, resource):
        if resource.get_resource_type() in self.map:
            raise Exception("Resource type {} already exists".format(resource.get_resource_type()))
        self.map[resource.get_resource_type()] = resource



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from enum import Enum
from typing import Dict

from fabric.actor.core.apis.i_delegation import IDelegation
from fabric.actor.core.apis.i_slice import ISlice
from fabric.actor.core.apis.i_kernel_reservation import IKernelReservation
from fabric.actor.core.apis.i_kernel_slice import IKernelSlice
from fabric.actor.core.util.id import ID
from fabric.actor.core.util.reservation_set import ReservationSet
from fabric.actor.core.util.resource_data import ResourceData
from fabric.actor.core.util.resource_type import ResourceType
from fabric.actor.security.auth_token import AuthToken
from fabric.actor.security.guard import Guard
from fim.graph.abc_property_graph import ABCPropertyGraph


class SliceTypes(Enum):
    InventorySlice = 1
    ClientSlice = 2
    BrokerClientSlice = 3


class Slice(IKernelSlice):
    """
    Slice implementation. A slice has a globally unique identifier, name,
    description, property list, an owning identity, an access control list, and a
    set of reservations.
    This class is used within the Service Manager, which may hold reservations on
    many sites; on the Broker, which may have provided tickets to the slice for
    reservations at many sites; and on the site Authority, where each slice may
    hold multiple reservations for resources at that site.
    """
    def __init__(self, *, id: ID = None, name: str = "unspecified", data: ResourceData = None):
        # Globally unique identifier.
        self.guid = id
        # Slice name. Not required to be globally or locally unique.
        self.name = name
        # Description string. Has only local meaning.
        self.description = "no description"
        # A collection of property lists inherited by each reservation in this
        # slice. Properties defined on the reservation level override properties
        # defined here.
        self.rsrcdata = data
        # The slice type: inventory or client.
        self.type = SliceTypes.ClientSlice
        # The owner of the slice.
        self.owner = None
        # Access control monitor.
        self.guard = Guard()
        # Resource type associated with this slice. Used when the slice is used to
        # represent an inventory pool.
        self.resource_type = None
        # The reservations in this slice.
        self.reservations = ReservationSet()
        self.delegations = {}
        # Neo4jGraph Id
        self.graph_id = None
        self.graph = None

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['reservations']
        del state['delegations']
        del state['graph']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.reservations = ReservationSet()
        self.graph = None
        self.delegations = {}

    def set_graph_id(self, graph_id: ID):
        self.graph_id = graph_id

    def get_graph_id(self) -> ID:
        return self.graph_id

    def set_graph(self, graph: ABCPropertyGraph):
        self.graph = graph
        self.set_graph_id(graph_id=ID(id=self.graph.get_graph_id()))

    def get_graph(self) -> ABCPropertyGraph:
        return self.graph

    def clone_request(self) -> ISlice:
        result = Slice()
        result.slice_name = self.name
        result.guid = self.guid
        return result

    def get_config_properties(self):
        if self.rsrcdata is not None:
            return self.rsrcdata.get_configuration_properties()
        return None

    def get_description(self):
        return self.description

    def get_guard(self) -> Guard:
        return self.guard

    def get_local_properties(self):
        if self.rsrcdata is not None:
            return self.rsrcdata.get_local_properties()
        return None

    def get_name(self):
        return self.name

    def get_owner(self):
        return self.owner

    def get_properties(self) -> ResourceData:
        return self.rsrcdata

    def get_request_properties(self):
        if self.rsrcdata is not None:
            return self.rsrcdata.get_request_properties()
        return None

    def get_reservations(self) -> ReservationSet:
        return self.reservations

    def get_delegations(self) -> Dict[str, IDelegation] :
        return self.delegations

    def get_reservations_list(self) -> list:
        return self.reservations.values()

    def get_resource_properties(self):
        if self.rsrcdata is not None:
            return self.rsrcdata.get_resource_properties()
        return None

    def get_resource_type(self):
        return self.resource_type

    def get_slice_id(self) -> ID:
        return self.guid

    def is_broker_client(self):
        return self.type == SliceTypes.BrokerClientSlice

    def is_client(self):
        return not self.is_inventory()

    def is_inventory(self):
        return self.type == SliceTypes.InventorySlice

    def is_empty(self) -> bool:
        return self.reservations.is_empty()

    def prepare(self):
        self.reservations.clear()

    def register(self, *, reservation: IKernelReservation):
        if self.reservations.contains(rid=reservation.get_reservation_id()):
            raise Exception("Reservation #{} already exists in slice".format(reservation.get_reservation_id()))

        self.reservations.add(reservation=reservation)

    def register_delegation(self, *, delegation: IDelegation):
        if delegation.get_delegation_id() in self.delegations:
            raise Exception("Delegation #{} already exists in slice".format(delegation.get_delegation_id()))

        self.delegations[delegation.get_delegation_id()] = delegation

    def set_broker_client(self):
        self.type = SliceTypes.BrokerClientSlice

    def set_client(self):
        self.type = SliceTypes.ClientSlice

    def set_description(self, *, description: str):
        self.description = description

    def set_guard(self, *, g: Guard):
        self.guard = g

    def set_inventory(self, *, value: bool):
        if value:
            self.type = SliceTypes.InventorySlice
        else:
            self.type = SliceTypes.ClientSlice

    def get_slice_type(self) -> SliceTypes:
        return self.type

    def set_name(self, *, name: str):
        self.name = name

    def set_owner(self, *, owner: AuthToken):
        self.owner = owner
        self.guard.set_owner(owner=owner)
        self.guard.set_object_id(object_id=self.guid)

    def set_properties(self, *, rsrcdata: ResourceData):
        self.rsrcdata = rsrcdata

    def set_resource_type(self, *, resource_type: ResourceType):
        self.resource_type = resource_type

    def soft_lookup(self, *, rid: ID) -> IKernelReservation:
        return self.reservations.get(rid=rid)

    def soft_lookup_delegation(self, *, did: ID) -> IDelegation:
        return self.delegations.get(did, None)

    def __str__(self):
        msg = "{}({})".format(self.name, str(self.guid))
        if self.graph_id is not None:
            msg += " Graph Id:{}".format(self.graph_id)
        return msg

    def unregister(self, *, reservation: IKernelReservation):
        self.reservations.remove(reservation=reservation)

    def unregister_delegation(self, *, delegation: IDelegation):
        if delegation.get_delegation_id() in self.delegations:
            self.delegations.pop(delegation.get_delegation_id())

    def set_local_properties(self, *, value: dict):
        if self.rsrcdata is not None:
            self.rsrcdata.local_properties = value

    def set_config_properties(self, *, value: dict):
        if self.rsrcdata is not None:
            self.rsrcdata.configuration_properties = value

    def set_request_properties(self, *, value: dict):
        if self.rsrcdata is not None:
            self.rsrcdata.request_properties = value

    def set_resource_properties(self, *, value: dict):
        if self.rsrcdata is not None:
            self.rsrcdata.resource_properties = value



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)

from fabric.actor.core.util.id import ID
from fabric.actor.security.credentials import Credentials


class AuthToken:
    PropertyName = 'name'
    PropertyGuid = 'guid'
    PropertyCred = 'cred'
    """
    Represents the Authentication Token for a user
    """
    def __init__(self, *, name: str = None, guid: ID = None, properties: dict = None):
        self.name = name
        self.guid = guid
        self.id_token = None
        self.cred = None

        if properties is not None:
            if self.PropertyName in properties:
                self.name = properties[self.PropertyName]
            if self.PropertyGuid in properties:
                self.guid = ID(id=properties[self.PropertyGuid])
            if self.PropertyCred in properties:
                self.cred = Credentials(properties=properties[self.PropertyCred])

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['cred']
        del state['id_token']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.cred = None
        self.id_token = None

    def get_name(self) -> str:
        return self.name

    def get_guid(self) -> ID:
        return self.guid

    def set_credentials(self, *, id_token: str):
        self.cred = Credentials(id_token=id_token)

    def get_credentials(self) -> Credentials:
        return self.cred




#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import logging

from fabric.actor.core.apis.i_actor import IActor, ActorType
from fabric.actor.core.apis.i_authority import IAuthority
from fabric.actor.core.apis.i_authority_policy import IAuthorityPolicy
from fabric.actor.core.apis.i_base_plugin import IBasePlugin
from fabric.actor.core.apis.i_broker import IBroker
from fabric.actor.core.apis.i_broker_policy import IBrokerPolicy
from fabric.actor.core.apis.i_container_database import IContainerDatabase
from fabric.actor.core.apis.i_controller import IController
from fabric.actor.core.apis.i_controller_policy import IControllerPolicy
from fabric.actor.core.apis.i_database import IDatabase
from fabric.actor.core.apis.i_policy import IPolicy
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.core.authority_policy import AuthorityPolicy
from fabric.actor.core.core.broker_policy import BrokerPolicy
from fabric.actor.core.core.policy import Policy
from fabric.actor.core.delegation.simple_resource_ticket_factory import SimpleResourceTicketFactory
from fabric.actor.core.plugins.base_plugin import BasePlugin
from fabric.actor.core.plugins.db.actor_database import ActorDatabase
from fabric.actor.core.policy.controller_calendar_policy import ControllerCalendarPolicy
from fabric.actor.core.registry.actor_registry import ActorRegistrySingleton
from fabric.actor.core.time.actor_clock import ActorClock
from fabric.actor.core.time.term import Term
from fabric.actor.core.util.id import ID
from fabric.actor.security.auth_token import AuthToken


class BaseTestCase:
    ActorName = "testActor"
    ActorGuid = ID()

    AuthorityName = "Test-Authority"
    BrokerName = "Test-Broker"
    ControllerName = "Test-Controller"

    ControllerGuid = "test-orchestrator-guid"
    AuthorityGuid = "test-authority-guid"
    BrokerGuid = "test-broker-guid"

    DbUser = 'fabric'
    DbPwd = 'fabric'
    DbName = 'test'
    DbHost = '127.0.0.1:8432'

    Logger = logging.getLogger('BaseTestCase')
    log_format = '%(asctime)s - %(name)s - {%(filename)s:%(lineno)d} - [%(threadName)s] - %(levelname)s - %(message)s'
    logging.basicConfig(format=log_format, filename="./actor.log")
    Logger.setLevel(logging.INFO)

    Term.set_cycles = False

    def get_container_database(self) -> IContainerDatabase:
        from fabric.actor.core.container.globals import GlobalsSingleton
        return GlobalsSingleton.get().get_container().get_database()

    def get_actor_clock(self) -> ActorClock:
        from fabric.actor.core.container.globals import GlobalsSingleton
        return GlobalsSingleton.get().get_container().get_actor_clock()

    def make_actor_database(self) -> IDatabase:
        return ActorDatabase(user=self.DbUser, password=self.DbPwd, database=self.DbName,
                             db_host=self.DbHost,
                             logger=self.Logger)

    def make_controller_database(self) -> IDatabase:
        return self.make_actor_database()

    def make_broker_database(self) -> IDatabase:
        return self.make_actor_database()

    def make_authority_database(self) -> IDatabase:
        return self.make_actor_database()

    def initialize_database(self, *, db: ActorDatabase, name: str):
        db.set_actor_name(name=name)
        db.initialize()

    def get_actor_database(self, *, name: str = None) -> IDatabase:
        if name is None:
            name = self.ActorName
        db = self.make_actor_database()
        self.initialize_database(db=db, name=name)
        return db

    def get_controller_database(self, *, name: str) -> IDatabase:
        db = self.make_controller_database()
        db.set_actor_name(name=name)
        db.initialize()
        return db

    def get_authority_database(self, *, name: str) -> IDatabase:
        db = self.make_authority_database()
        db.set_actor_name(name=name)
        db.initialize()
        return db

    def get_broker_database(self, *, name: str) -> IDatabase:
        db = self.make_broker_database()
        db.set_actor_name(name=name)
        db.initialize()
        return db

    def make_plugin(self) -> IBasePlugin:
        return BasePlugin(actor=None, db=None, config=None)

    def get_plugin(self, *, name: str = None) -> IBasePlugin:
        if name is None:
            name = self.ActorName
        plugin = self.make_plugin()
        plugin.set_database(db=self.get_actor_database(name=name))
        return plugin

    def get_controller_plugin(self, *, name: str) -> IBasePlugin:
        plugin = self.make_plugin()
        plugin.set_database(db=self.get_controller_database(name=name))
        return plugin

    def get_broker_plugin(self, *, name: str) -> IBasePlugin:
        plugin = self.make_plugin()
        plugin.set_database(db=self.get_broker_database(name=name))
        return plugin

    def get_authority_plugin(self, *, name: str) -> IBasePlugin:
        plugin = self.make_plugin()
        plugin.set_database(db=self.get_authority_database(name=name))
        return plugin

    def get_policy(self) -> IPolicy:
        return Policy()

    def get_authority_policy(self) -> IAuthorityPolicy:
        return AuthorityPolicy(actor=None)

    def get_broker_policy(self) -> IBrokerPolicy:
        return BrokerPolicy(actor=None)

    def get_controller_policy(self) -> IControllerPolicy:
        return ControllerCalendarPolicy()

    def get_actor_instance(self) -> IActor:
        from fabric.actor.test.test_actor import TestActor
        actor = TestActor()
        actor.type = ActorType.All
        return actor

    def get_authority_instance(self) -> IActor:
        from fabric.actor.core.core.authority import Authority
        return Authority()

    def get_broker_instance(self) -> IActor:
        from fabric.actor.core.core.broker import Broker
        return Broker()

    def get_controller_instance(self) -> IActor:
        from fabric.actor.core.core.controller import Controller
        return Controller()

    def get_uninitialized_actor(self, *, name: str, guid: ID) -> IActor:
        actor = self.get_actor_instance()
        token = AuthToken(name=name, guid=guid)
        actor.set_identity(token=token)
        actor.set_actor_clock(clock=self.get_actor_clock())
        actor.set_policy(policy=self.get_policy())
        actor.set_plugin(plugin=self.get_plugin(name=name))
        tf = SimpleResourceTicketFactory()
        tf.set_actor(actor=actor)
        actor.get_plugin().set_ticket_factory(ticket_factory=tf)
        return actor

    def get_uninitialized_controller(self, *, name: str, guid: ID) -> IController:
        actor = self.get_controller_instance()
        token = AuthToken(name=name, guid=guid)
        actor.set_identity(token=token)
        actor.set_actor_clock(clock=self.get_actor_clock())
        actor.set_policy(policy=self.get_controller_policy())
        actor.set_plugin(plugin=self.get_plugin(name=name))
        tf = SimpleResourceTicketFactory()
        tf.set_actor(actor=actor)
        actor.get_plugin().set_ticket_factory(ticket_factory=tf)
        return actor

    def get_uninitialized_broker(self, *, name: str, guid: ID) -> IBroker:
        actor = self.get_broker_instance()
        token = AuthToken(name=name, guid=guid)
        actor.set_identity(token=token)
        actor.set_actor_clock(clock=self.get_actor_clock())
        actor.set_policy(policy=self.get_broker_policy())
        actor.set_plugin(plugin=self.get_plugin(name=name))
        tf = SimpleResourceTicketFactory()
        tf.set_actor(actor=actor)
        actor.get_plugin().set_ticket_factory(ticket_factory=tf)
        return actor

    def get_uninitialized_authority(self, *, name: str, guid: ID) -> IAuthority:
        actor = self.get_authority_instance()
        token = AuthToken(name=name, guid=guid)
        actor.set_identity(token=token)
        actor.set_actor_clock(clock=self.get_actor_clock())
        actor.set_policy(policy=self.get_authority_policy())
        actor.set_plugin(plugin=self.get_plugin(name=name))
        tf = SimpleResourceTicketFactory()
        tf.set_actor(actor=actor)
        actor.get_plugin().set_ticket_factory(ticket_factory=tf)
        return actor

    def get_actor(self, *, name: str = ActorName, guid: ID = ActorGuid) -> IActor:
        actor = self.get_uninitialized_actor(name=name, guid=guid)
        actor.initialize()
        self.register_new_actor(actor=actor)
        return actor

    def get_controller(self, *, name: str = ControllerName, guid: ID = ControllerGuid) -> IController:
        actor = self.get_uninitialized_controller(name=name, guid=guid)
        actor.initialize()
        self.register_new_actor(actor=actor)
        return actor

    def get_broker(self, *, name: str = BrokerName, guid: ID = BrokerGuid) -> IBroker:
        actor = self.get_uninitialized_broker(name=name, guid=guid)
        actor.initialize()
        self.register_new_actor(actor=actor)
        return actor

    def get_authority(self, *, name: str = AuthorityName, guid: ID = AuthorityGuid) -> IAuthority:
        actor = self.get_uninitialized_authority(name=name, guid=guid)
        actor.initialize()
        self.register_new_actor(actor=actor)
        return actor

    def register_new_actor(self, *, actor: IActor):
        db = self.get_container_database()
        db.remove_actor(actor_name=actor.get_name())
        db.add_actor(actor=actor)
        ActorRegistrySingleton.get().unregister(actor=actor)
        ActorRegistrySingleton.get().register_actor(actor=actor)
        actor.actor_added()
        actor.start()

    def get_registered_new_actor(self) -> IActor:
        actor = self.get_actor()
        self.register_new_actor(actor=actor)
        return actor


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

from fabric.actor.core.apis.i_client_reservation import IClientReservation
from fabric.actor.core.apis.i_slice import ISlice

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_substrate import ISubstrate
    from fabric.actor.core.common.resource_pool_descriptor import ResourcePoolDescriptor


class IResourcePoolFactory:
    @abstractmethod
    def set_substrate(self, *, substrate: ISubstrate):
        """
        Sets the substrate.
        @param substrate substrate
        @throws Exception in case of error
        """

    @abstractmethod
    def set_descriptor(self, *, descriptor: ResourcePoolDescriptor):
        """
        Sets the initial resource pools descriptor. The factory can modify the descriptor as needed.
        @param descriptor descriptor
        @throws Exception in case of error
        """

    @abstractmethod
    def get_descriptor(self) -> ResourcePoolDescriptor:
        """
        Returns the final pool descriptor.
        @return ResourcePoolDescriptor
        @throws Exception in case of error
        """

    @abstractmethod
    def create_source_reservation(self, *, slice_obj: ISlice) -> IClientReservation:
        """
        Returns the source reservation for this resource pool.
        @param slice_obj slice
        @return IClientReservation
        @throws Exception in case of error
        """



import traceback

import connexion
import six

from fabric.orchestrator.core.orchestrator_handler import OrchestratorHandler
from fabric.orchestrator.swagger_server.models.success import Success  # noqa: E501
from fabric.orchestrator.swagger_server import util, received_counter, success_counter, failure_counter


def resources_get():  # noqa: E501
    """Retrieve a listing and description of available resources

    Retrieve a listing and description of available resources # noqa: E501


    :rtype: Success
    """
    handler = OrchestratorHandler()
    logger = handler.get_logger()
    received_counter.labels('get', '/resources').inc()
    try:
        value = handler.list_resources()
        response = Success()
        response.value = value
        success_counter.labels('get', '/resources').inc()
        return response
    except Exception as e:
        logger.exception(e)
        failure_counter.labels('get', '/resources').inc()
        return str(e), 500




#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations


from typing import TYPE_CHECKING

from fabric.actor.core.common.constants import Constants

if TYPE_CHECKING:
    from fabric.actor.core.kernel.resource_set import ResourceSet


class PropList:
    @staticmethod
    def merge_properties(*, incoming: dict, outgoing: dict):
        if incoming is None:
            return outgoing

        if outgoing is None:
            return incoming

        if incoming == outgoing:
            return incoming

        if incoming is not None and outgoing is not None:
            return {**incoming, **outgoing}

    @staticmethod
    def is_elastic_time(*, rset: ResourceSet):
        properties = rset.get_request_properties()
        if properties is not None and Constants.ElasticTime in properties:
            value = properties[Constants.ElasticTime]
            if value.lower() == 'true':
                return True

        return False

    @staticmethod
    def is_elastic_size(*, rset: ResourceSet):
        properties = rset.get_request_properties()
        if properties is not None and Constants.ElasticSize in properties:
            value = properties[Constants.ElasticSize]
            if value.lower() == 'true':
                return True

        return False


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from enum import Enum


class ErrorCodes(Enum):
    ErrorInvalidArguments = 1
    ErrorInvalidActor = 2
    ErrorInvalidReservation = 3
    ErrorDatabaseError = 4
    ErrorInternalError = 5
    ErrorNoSuchSlice = 6
    ErrorNoSuchResourcePool = 7
    ErrorNoSuchReservation = 8
    ErrorNoSuchBroker = 9
    ErrorInvalidSlice = 10
    ErrorNoSuchActor = 11
    ErrorTransportFailure = 12
    ErrorTransportTimeout = 13


class Constants:
    ReservationHasPendingOperation = -200001
    ExtendSameUnits = -1
    AllReservationStates = -1
    ManagementApiTimeoutInSeconds = 120.0

    ContainerManagmentObjectID = "manager"

    PropertyPickleProperties = "properties"

    ConfigSectionRuntime = "runtime"
    PropertyConfKafkaServer = "kafka-server"
    PropertyConfKafkaSchemaRegistry = "kafka-schema-registry-url"
    PropertyConfKafkaKeySchema = "kafka-key-schema"
    PropertyConfKafkaValueSchema = "kafka-value-schema"
    PropertyConfKafkaSSlCaLocation = "kafka-ssl-ca-location"
    PropertyConfKafkaSslCertificateLocation = "kafka-ssl-certificate-location"
    PropertyConfKafkaSslKeyLocation = "kafka-ssl-key-location"
    PropertyConfKafkaSslKeyPassword = "kafka-ssl-key-password"
    PropertyConfKafkaSecurityProtocol = "kafka-security-protocol"
    PropertyConfKafkaGroupId = "kafka-group-id"
    PropertyConfKafkaSaslProducerUsername = "kafka-sasl-producer-username"
    PropertyConfKafkaSaslProducerPassword = "kafka-sasl-producer-password"
    PropertyConfKafkaSaslConsumerUsername = "kafka-sasl-consumer-username"
    PropertyConfKafkaSaslConsumerPassword = "kafka-sasl-consumer-password"
    PropertyConfKafkaSaslMechanism = "kafka-sasl-mechanism"

    PropertyConfPluginDir = "plugin-dir"

    ConfigLoggingSection = 'logging'

    PropertyConfLogFile = 'log-file'
    PropertyConfLogLevel = 'log-level'
    PropertyConfLogRetain = 'log-retain'
    PropertyConfLogSize = 'log-size'
    PropertyConfLogDirectory = 'log-directory'
    PropertyConfLogger = "logger"

    ConfigSectionContainer = "container"
    PropertyConfContainerGuid = "container.guid"

    ConfigSectionTime = "time"
    PropertyConfTimeStartTime = "time.startTime"
    PropertyConfTimeCycleMillis = "time.cycleMillis"
    PropertyConfTimeManual = "time.manual"

    ConfigSectionOAuth = "oauth"
    PropertyConfOAuthJwksUrl = "oauth-jwks-url"
    PropertyConfOAuthTokenPublicKey = "token-public-key"

    ConfigSectionDatabase = "database"
    PropertyConfDbUser = "db-user"
    PropertyConfDbPassword = "db-password"
    PropertyConfDbName = "db-name"
    PropertyConfDbHost = "db-host"

    ConfigSectionNeo4j = "neo4j"

    # Type code for proxies using local communication.
    ProtocolLocal = "local"
    # Type code for proxies using Kafka communication.
    ProtocolKafka = "kafka"

    HomeDirectory = '/usr/src/app/'
    SuperblockLocation = HomeDirectory + "state_recovery.lock"
    ControllerLockLocation = HomeDirectory + "controller_recovery.lock"
    ConfigurationFile = "/etc/fabric/actor/config/config.yaml"
    StateFileLocation = '/tmp/fabric_actor.tmp'

    TestDirectory = "/Users/komalthareja/renci/code/fabric/ControlFramework/fabric/actor/test/"
    TestBrokerConfigurationFile = TestDirectory + "config/config.broker.yaml"
    TestNetAmConfigurationFile = TestDirectory + "config/config.net-am.yaml"
    TestVmAmConfigurationFile = TestDirectory + "config/config.site.am.yaml"
    TestControllerConfigurationFile = TestDirectory + "config/config.orchestrator.yaml"

    ElasticTime = "request.elasticTime"
    ElasticSize = "request.elasticSize"

    PoolName = 'pool.name'
    PoolPrefix = "pool."
    PoolsCount = "pools.count"
    QueryActionDiscoverPools = "discover.pools"
    QueryAction = "query.action"
    QueryResponse = "query.response"
    BrokerQueryModel = "bqm"
    PoolType = "neo4j"

    ConfigHandler = "config.handler"
    ConfigVictims = "config.victims"
    ConfigImageGuid = "config.image.guid"
    ConfigSSHKeyPattern = "config.ssh.user%d.keys"
    ConfigSSHLoginPattern = "config.ssh.user%d.login"
    ConfigSSHSudoPattern = "config.ssh.user%d.sudo"
    ConfigSSHUrnPattern = "config.ssh.user%d.urn"
    ConfigSSHNumLogins = "config.ssh.numlogins"
    ConfigSSHPrefix = "config.ssh.user"
    ConfigSSHKeySuffix = ".keys"
    ConfigSSHLoginSuffix = ".login"
    ConfigSSHSudoSuffix = ".sudo"
    ConfigSSHUrnSuffix = ".urn"

    ResourceMemory = "resource.memory"
    ResourceCPU = "resource.cpu"
    ResourceBandwidth = "resource.bandwidth"
    ResourceNumCPUCores = "resource.numCPUCores"
    ResourceMemoryCapacity = "resource.memeoryCapacity"
    ResourceStorageCapacity = "resource.storageCapacity"
    ResourceClassInventoryForType = "resource.class.invfortype"
    ResourceAvailableUnits = "resource.units.now"
    ResourceStartIface = "resource.siface"
    ResourceEndIface = "resource.eiface"
    ResourceDomain = "resource.domain"
    ResourceNeo4jAbstractDomain = "resource.neo4j.adomain"

    UnitManagementIP = "unit.manage.ip"
    UnitManagementPort = "unit.manage.port"
    UnitManageSubnet = "unit.manage.subnet"
    UnitManageGateway = "unit.manage.gateway"

    UnitDataSubnet = "unit.data.subnet"

    UnitParentHostName = "unit.parent.hostname"
    UnitHostName = "unit.hostname"
    UnitControl = "unit.control"
    UnitMemory = "unit.memory"

    UnitVlanTag = "unit.vlan.tag"
    UnitVlanQoSRate = "unit.vlan.qos.rate"
    UnitVlanQoSBurstSize = "unit.vlan.qos.burst.size"

    UnitEthPrefix = "unit.eth"
    UnitNumberInterface = "unit.number.interface"

    UnitModifyPropMessageSuffix = ".message"
    UnitModifyPropCodeSuffix = ".code"
    UnitModifyPropPrefix = "unit.modify."

    ConfigUnitTag = "config.unit.tag"

    PropertyStartVlan = "vlan.tag.start"
    PropertyEndVlan = "vlan.tag.end"
    PropertyVlanRangeNum = "vlan.range.num"

    PropertyStartLUN = "lun.tag.start"
    PropertyEndLUN = "lun.tag.end"
    PropertyLunRangeNum = "lun.range.num"

    UnitLUNTag = "unit.target.lun"
    UnitStorageCapacity = "unit.target.capacity"

    KafkaTopic = "kafka-topic"
    Name = "name"
    Type = "type"
    Guid = "guid"

    PropertyClassName = "ObjectClassName"
    PropertyModuleName = "ModuleName"
    PropertyID = "MOID"
    PropertyTypeID = "MOTYPEID"
    PropertyActorName = "MOActorName"
    PropertyProxiesLength = "MOProxiesLength"
    PropertyProxiesPrefix = "MOProxiesPrefix."
    PropertyProxiesProtocol = ".protocol"
    PropertyProxiesClass = ".class"
    PropertyProxiesModule = ".module"

    PropertyConfPrometheusRestPort = "prometheus.port"
    PropertyConfControllerRestPort = "orchestrator.rest.port"
    PropertyConfControllerCreateWaitTimeMs = "orchestrator.create.wait.time.ms"
    PropertyConfControllerDelayResourceTypes = "orchestrator.delay.resource.types"

    PropertySubstrateFile = "substrate.file"
    PropertyAggregateResourceModel = "AggregateResourceModel"

    PropertyDelegationSliceId = 'dlg_slc_id'
    PropertyReservationSliceId = 'rsv_slc_id'
    PropertyReservationID = "rsv_resid"


import connexion
import requests
import six

from fabric.orchestrator.swagger_server.models.version import Version  # noqa: E501
from fabric.orchestrator.swagger_server import util, received_counter, success_counter, failure_counter


def version_get():  # noqa: E501
    """version

    Version # noqa: E501


    :rtype: Version
    """
    received_counter.labels('get', '/version').inc()
    from fabric.actor.core.container.globals import GlobalsSingleton
    logger = GlobalsSingleton.get().get_logger()
    try:
        version = '1.0.0'
        tag = '1.0.0'
        url = "https://api.github.com/repos/fabric-testbesd/ControlFramework/git/refs/tags/{}".format(tag)

        response = Version()
        response.version = version
        response.gitsha1 = 'Not Available'

        result = requests.get(url)
        if result.status_code == 200:
            if result.json() is not None:
                object_json = result.json().get("object", None)
                if object_json is not None:
                    sha = object_json.get("sha", None)
                    if sha is not None:
                        response.gitsha1 = sha
        success_counter.labels('get', '/version').inc()
    except Exception as e:
        logger.exception(e)
        failure_counter.labels('get', '/version').inc()
        return str(e), 500
    return response



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from enum import Enum


class ResourcePoolAttributeType(Enum):
    UNDEFINED = 1
    INTEGER = 2
    FLOAT = 3
    STRING = 4
    NEO4J = 5
    CLASS = 6


class ResourcePoolAttributeDescriptor:
    PropertyType = "type"
    PropertyUnit = "unit"
    PropertyValue = "value"
    PropertyLabel = "label"

    def __init__(self):
        self.key = None
        self.type = None
        self.value = None
        self.label = None
        self.unit = None

    def get_type(self) -> ResourcePoolAttributeType:
        return self.type

    def set_type(self, *, rtype: ResourcePoolAttributeType):
        self.type = rtype

    def get_value(self) -> str:
        return self.value

    def get_int_value(self) -> int:
        return int(self.value)

    def set_value(self, *, value: str):
        self.value = value

    def get_key(self) -> str:
        return self.key

    def set_key(self, *, value: str):
        self.key = value

    def set_label(self, *, label: str):
        self.label = label

    def get_label(self) -> str:
        return self.label

    def get_unit(self) -> str:
        return self.unit

    def set_unit(self, *, unit: str):
        self.unit = unit

    def __str__(self):
        return "{}:{}:{}".format(self.key, self.value, self.type)

    def save(self, *, properties: dict, prefix: str) -> dict:
        if prefix is None:
            prefix = ""

        properties[prefix + self.PropertyType] = str(self.type.value)
        if self.value is not None:
            properties[prefix + self.PropertyValue] = self.value

        if self.label is not None:
            properties[prefix + self.PropertyLabel] = self.label

        if self.unit is not None:
            properties[prefix + self.PropertyUnit] = self.unit

        return properties

    def reset(self, *, properties: dict, prefix: str):
        if prefix is None:
            prefix = ""

        self.type = ResourcePoolAttributeType(int(properties[prefix + self.PropertyType]))
        self.value = properties[prefix + self.PropertyValue]
        if prefix + self.PropertyLabel in properties:
            self.label = properties[prefix + self.PropertyLabel]
        if prefix + self.PropertyUnit in properties:
            self.unit = properties[prefix + self.PropertyUnit]



from typing import List

from fabric.orchestrator.core.orchestrator_handler import OrchestratorHandler

"""
controller generated to handled auth operation described at:
https://connexion.readthedocs.io/en/latest/security.html
"""
def check_bearerAuth(token):
    orchestrator = OrchestratorHandler()
    return orchestrator.validate_credentials(token=token)





#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from datetime import datetime
from typing import TYPE_CHECKING, List

from fabric.actor.core.apis.i_mgmt_actor import IMgmtActor
from fim.graph.abc_property_graph import ABCPropertyGraph

if TYPE_CHECKING:
    from fabric.actor.core.util.id import ID
    from fabric.actor.core.manage.messages.client_mng import ClientMng
    from fabric.actor.security.auth_token import AuthToken
    from fabric.actor.core.util.resource_type import ResourceType
    from fabric.message_bus.messages.reservation_mng import ReservationMng
    from fabric.message_bus.messages.slice_avro import SliceAvro

class IMgmtServerActor(IMgmtActor):
    @abstractmethod
    def get_broker_reservations(self) -> List[ReservationMng]:
        """
        Returns all reservations for which this actor acts as a broker.
        @return list of the reservations
        """

    @abstractmethod
    def get_inventory_slices(self) -> List[SliceAvro]:
        """
        Obtains all slices holding inventory, i.e., resources that can be
        delegated to other actors.
        @return list of slices
        """

    @abstractmethod
    def get_inventory_reservations(self) -> List[ReservationMng]:
        """
        Returns all resources held by this actor that can be used for delegations
        to client actors.
        @return list of reservations
        """

    @abstractmethod
    def get_inventory_reservations_by_slice_id(self, *, slice_id: ID) -> List[ReservationMng]:
        """
        Returns all resources in the specified slice held by this actor that can
        be used for delegations to client actors.
        @param sliceID slice id
        @return list of reservations for specific slice
        """

    @abstractmethod
    def get_client_slices(self) -> List[SliceAvro]:
        """
        Obtains all slices that hold delegated resources to other actors.
        @return list of client slices
        """

    @abstractmethod
    def add_client_slice(self, *, slice_mng: SliceAvro) -> ID:
        """
        Adds a new client slice.
        @param slice_mng slice to be added
        @return sliceid of the added slice
        """

    @abstractmethod
    def get_clients(self) -> List[ClientMng]:
        """
        Returns all registered clients of this server actor.
        @return list of clients
        """

    @abstractmethod
    def get_client(self, *, guid: ID) -> ClientMng:
        """
        Returns the specified client record.
        @param guid client guid
        @return specified client record
        """

    @abstractmethod
    def register_client(self, *, client: ClientMng, kafka_topic: str) -> bool:
        """
        Registers a new client
        @param client client
        @param kafka_topic Kafka topic
        @return true for success; false otherwise
        """

    @abstractmethod
    def unregister_client(self, *, guid: ID) -> bool:
        """
        Unregisters the specified client.
        @param guid client guid
        @return true for success; false otherwise
        """

    @abstractmethod
    def get_client_reservations(self) -> List[ReservationMng]:
        """
        Obtains all client reservations.
        @return list of client reservations
        """

    @abstractmethod
    def get_client_reservations_by_slice_id(self, *, slice_id: ID) -> List[ReservationMng]:
        """
        Obtains all client reservations in the specified slice
        @param slice_id slice id
        @return list of reservations
        """

    @abstractmethod
    def export_resources_pool_client_slice(self, *, client_slice_id: ID, pool_id: ID, start: datetime, end: datetime,
                                           units: int, ticket_properties: dict, resource_properties: dict,
                                           source_ticket_id: ID) -> ID:
        """
        Exports resources into the specified client slice from the specified
        resource pool using the given source reservation. units number
        of units are exported from start to end.
        All properties passed into ticketProperties will be part of the ticket and signed.
        All properties passed into resourceProperties will be attached as resource propertie
        to the resource set (unsigned).
        @param client_slice_id client slice id
        @param pool_id pool slice id
        @param start start date
        @param end end date
        @param units units
        @param ticket_properties ticket properties
        @param resource_properties resource properties
        @param source_ticket_id ticket id
        @return returns the reservation id
        """

    @abstractmethod
    def export_resources_pool(self, *, pool_id: ID, start: datetime, end: datetime, units: int,
                              ticket_properties: dict, resource_properties: dict, source_ticket_id: ID,
                              client: AuthToken) -> ID:
        """
        Exports resources  from the specified
        resource pool using the given source reservation. units number
        of units are exported from start to end.
        All properties passed into ticketProperties will be part of the ticket and signed.
        All properties passed into resourceProperties will be attached as resource propertie
        to the resource set (unsigned).
        @param pool_id pool slice id
        @param start start date
        @param end end date
        @param units units
        @param ticket_properties ticket properties
        @param resource_properties resource properties
        @param source_ticket_id ticket id
        @param client client
        @return returns the reservation id
        """
    @abstractmethod
    def export_resources_client_slice(self, *, client_slice_id: ID, rtype: ResourceType, start: datetime, end: datetime,
                                      units: int, ticket_properties: dict, resource_properties: dict,
                                      source_ticket_id: ID) -> ID:
        """
        Exports resources into the specified client slice using the given source reservation. units number
        of units are exported from start to end.
        All properties passed into ticketProperties will be part of the ticket and signed.
        All properties passed into resourceProperties will be attached as resource propertie
        to the resource set (unsigned).
        @param client_slice_id client slice id
        @param rtype rtype
        @param start start date
        @param end end date
        @param units units
        @param ticket_properties ticket properties
        @param resource_properties resource properties
        @param source_ticket_id ticket id
        @return returns the reservation id
        """

    @abstractmethod
    def export_resources(self, *, rtype: ResourceType, start: datetime, end: datetime, units: int,
                         ticket_properties: dict, resource_properties: dict, source_ticket_id: ID,
                         client: AuthToken) -> ID:
        """
        Exports resources into the specified client slice from the specified
        resource pool using the given source reservation. units number
        of units are exported from start to end.
        All properties passed into ticketProperties will be part of the ticket and signed.
        All properties passed into resourceProperties will be attached as resource propertie
        to the resource set (unsigned).
        @param rtype rtype
        @param start start date
        @param end end date
        @param units units
        @param ticket_properties ticket properties
        @param resource_properties resource properties
        @param source_ticket_id ticket id
        @param client client
        @return returns the reservation id
        """

    @abstractmethod
    def advertise_resources(self, *, delegation: ABCPropertyGraph, client: AuthToken) -> ID:
        """
        Advertise resources to the broker
        @param delegation: delegation
        @param client: client
        @return return the reservation id
        """


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import datetime

from fabric.actor.core.apis.i_reservation import IReservation
from fabric.actor.core.util.id import ID
from fabric.actor.core.util.resource_count import ResourceCount


class ReservationSet:
    """
    ReservationSet is a collection of reservations indexed by ReservationID
    """
    def __init__(self, *, reservations: dict = None):
        self.reservations = {}
        if reservations is not None:
            self.reservations = reservations

    def __str__(self):
        result = ""
        for r in self.reservations.values():
            result += "rid={} r={}".format(r.get_reservation_id(), r)
        return result

    def add(self, *, reservation: IReservation):
        """
        Adds the reservation to the set

        Args:
            reservation: reservation to be added
        """
        self.reservations[reservation.get_reservation_id()] = reservation

    def clear(self):
        """
        Remove all the reservations from the set
        """
        self.reservations.clear()

    def contains(self, *, reservation: IReservation = None, rid: ID = None):
        """
        Checks if the reservation is part of the set

        Args:
            reservation: reservation to check
            rid: reservation id
        Returns:
            true if the set contains the specified reservation; false otherwise
        """
        if reservation is not None and reservation.get_reservation_id() in self.reservations :
            return True
        if rid is not None and rid in self.reservations:
            return True
        return False

    def count(self, *, rc: ResourceCount, when: datetime):
        """
        Tallies up resources in the ReservationSet. Note: "just a hint" unless kernel lock is held.

        Args:
            rc: holder for counts
            when: date relative to which to do the counting
        """
        for reservation in self.reservations.values():
            reservation.count(rc, when)

    def get(self, *, rid: ID) -> IReservation:
        """
        Retrieves a reservation from the set.

        Args:
            rid: reservation id
        Returns:
            Reservation identified by rid
        """
        return self.reservations.get(rid, None)

    def get_exception(self, *, rid: ID) -> IReservation:
        """
        Returns the specified reservation. If the reservation is not
        present in the set, throws an exception.
       
        @param rid the reservation identifier
       
        @return Reservation identified by rid
       
        @throws Exception if the requested reservation is not present in the set
        """
        if rid in self.reservations:
            return self.reservations.get(rid)

        raise Exception("No reservation with ID {}".format(rid))

    def is_empty(self) -> bool:
        """
        Checks if the set is empty.

        Returns:
            true if the set is empty
        """
        if len(self.reservations.keys()) == 0:
            return True
        return False

    def remove(self, *, reservation: IReservation):
        """
        Removes the specified reservation.

        Args:
            reservation: reservation to remove
        """
        if reservation.get_reservation_id() in self.reservations:
            self.reservations.pop(reservation.get_reservation_id())

    def remove_by_rid(self, *, rid: ID):
        """
        Removes the specified reservation.

        Args:
            rid: reservation id of reservation to remove
        """
        if rid in self.reservations:
            self.reservations.pop(rid)

    def size(self) -> int:
        """
        Returns the number of reservations in the set.

        Returns:
            the size of the reservation set
        """
        return len(self.reservations.keys())

    def __eq__(self, other):
        if not isinstance(other, ReservationSet):
            # don't attempt to compare against unrelated types
            return NotImplemented

        return self.reservations == other.reservations

    def clone(self):
        result = ReservationSet()
        result.reservations = self.reservations.copy()
        return result

    def values(self) -> list:
        result = []
        for r in self.reservations.values():
            result.append(r)
        return result


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import threading
from typing import List

from fabric.actor.core.apis.i_mgmt_actor import IMgmtActor
from fabric.actor.core.util.id import ID
from fabric.orchestrator.core.active_status_checker import ActiveStatusChecker
from fabric.orchestrator.core.i_status_update_callback import IStatusUpdateCallback
from fabric.orchestrator.core.modify_status_checker import ModifyStatusChecker
from fabric.orchestrator.core.status_checker import StatusChecker, Status
from fabric.orchestrator.core.watch_entry import WatchEntry, TriggeredWatchEntry


class ReservationStatusUpdateThread:
    """
    This thread allows expressing interest in completion of certain reservations and can run callbacks on other specified
    reservations when the status changes accordingly

    Supports modify and more traditional ticketed-active

    Principle of operation: allows to watch the following state transitions and events: - to Active transitions on
    reservations - to Active transitions followed by OK unit status on modify

    The periodic thread polls reservations of interest to determine whether the they have reached the required state,
    upon which
    """
    MODIFY_CHECK_PERIOD = 5 # seconds

    def __init__(self):
        self.lock = threading.Lock()
        self.active_watch = []
        self.modify_watch = []
        from fabric.actor.core.container.globals import GlobalsSingleton
        self.logger = GlobalsSingleton.get().get_logger()

        self.thread = None
        self.stopped_worker = threading.Event()

    def start(self):
        self.thread = threading.Thread(target=self.periodic())
        self.thread.setName('ReservationStatusUpdateThread')
        self.thread.setDaemon(True)
        self.thread.start()

    def stop(self):
        self.stopped_worker.set()
        self.thread.join()

    def periodic(self):
        while not self.stopped_worker.wait(timeout=self.get_period()):
            self.run()

    def add_active_status_watch(self, *, watch: List[ID], act: List[ID], callback: IStatusUpdateCallback):
        """
        Watch for transition to Active or Failed. Callback is called when ALL reservations in the watch list have reached
        Active or Failed or Closed state. If all reservations on the watch list went to Active or Closed, the success
        method of the callback is called If some or all reservations on the watch list went to Failed state, the failure
        method of the callback is called.
        @param watch - reservations to watch
        @param act - reservations to act on (can be null)
        @param callback - callback object
        """
        if watch is None or len(watch) == 0 or callback is None:
            self.logger.info("watch list is size 0 or callback is null, ignoring")
            return

        try:
            self.lock.acquire()
            self.active_watch.append(WatchEntry(watch=watch, rids=act, callback=callback))
        finally:
            self.lock.release()

    def add_modify_status_watch(self, *, watch, act, callback: IStatusUpdateCallback):
        """
        Watch for OK unit modify status (or not OK). Callback is called when ALL reservations in the watch list have
        reached some sort of modify status (OK or not OK) If all reservations on the watch list went to OK, the success
        method of the callback is called If some or all reservations on the watch list went to not OK, the failure method
        of the callback is called.
        @param watch - reservations to watch (with associated modify operation indices)
        @param act - reservations to act on (can be null)
        @param callback - callback
        """
        if watch is None or len(watch) == 0 or callback is None:
            self.logger.info("watch list is size 0 or callback is null, ignoring")
            return

        try:
            self.lock.acquire()
            self.modify_watch.append(WatchEntry(watch=watch, rids=act, callback=callback))
        finally:
            self.lock.release()

    def check_watch_entry(self, *, controller: IMgmtActor, watch_entry: WatchEntry,
                          status_checker: StatusChecker) -> TriggeredWatchEntry:

        ok = []
        notok = []

        ready = True
        for rid in watch_entry.watch:
            status = status_checker.check(controller=controller, rid=rid, ok=ok, not_ok=notok)
            if status == Status.NOTREADY:
                ready = False

        if not ready:
            self.logger.debug("Reservation watch not ready for reservations {}".format(watch_entry.watch))
            return None

        return TriggeredWatchEntry(watch=watch_entry.watch, rids=watch_entry.act, callback=watch_entry.callback, ok=ok,
                                   no_ok=notok)

    def process_callback(self, *, watch_entry: TriggeredWatchEntry):
        if len(watch_entry.not_ok) == 0:
            self.logger.debug("Invoking success callback for reservations: {}".format(watch_entry.watch))
            watch_entry.callback.success(watch_entry.ok, watch_entry.act)
        else:
            self.logger.debug("Invoking failure callback for reservations: {}".format(watch_entry.watch))
            watch_entry.callback.failure(watch_entry.not_ok, watch_entry.ok, watch_entry.act)

    def process_watch_list(self, *, controller: IMgmtActor, watch_list: List[WatchEntry], watch_type: str,
                           status_checker: StatusChecker):
        to_remove = []
        self.logger.debug("Scanning {} watch list".format(watch_type))
        to_process = []
        try:
            self.lock.acquire()
            for watch_entry in watch_list:
                twe = self.check_watch_entry(controller=controller, watch_entry=watch_entry,
                                             status_checker=status_checker)
                if twe is not None:
                    to_process.append(twe)
                    to_remove.append(twe)
            self.logger.debug("Removing {} entries from watch {}".format(watch_type, len(to_remove)))

            for we in to_remove:
                watch_list.remove(we)
        finally:
            self.lock.release()

        self.logger.debug("Processing {} triggered {} callbacks".format(len(to_process), watch_type))
        for we in to_process:
            try:
                self.process_callback(watch_entry=we)
            except Exception as e:
                self.logger.error("Triggered {} watch entry for reservations {} returned with callback exception e: {}".
                                  format(watch_type, we.watch, e))

    def run(self):
        # wake up periodically and check the status of reservations (state or unit modify properties)
        # and call appropriate callbacks if necessary.
        # scan both lists and check if any of the
        # reservation groups on them are ready for callbacks
        # remove those ready for callbacks off the lists and invoke callbacks
        # outside the critical section
        # NOTE: because callback can create another callback, which can add watches,
        # we must make this code re-entrant hence the need to change the watch list
        # and only then call the callbacks

        try:
            from fabric.orchestrator.core.orchestrator_state import OrchestratorStateSingleton
            controller = OrchestratorStateSingleton.get().get_management_actor()
            self.process_watch_list(controller=controller, watch_list=self.active_watch, watch_type="active",
                                    status_checker=ActiveStatusChecker())
            self.process_watch_list(controller=controller, watch_list=self.modify_watch, watch_type="modify",
                                    status_checker=ModifyStatusChecker())
        except Exception as e:
            self.logger.error("RuntimeException: {} continuing".format(e))

    @staticmethod
    def get_period() -> int:
        return ReservationStatusUpdateThread.MODIFY_CHECK_PERIOD



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric.actor.security.auth_token import AuthToken
from fabric.actor.security.guard import Guard


class AccessMonitor:
    """
    AccessMonitor encapsulates access control policy and any means of proving or gaining access to resources,
    e.g., payment. This should be an abstract pluggable class. Each operation on a slices actor includes an AuthToken
    to represent identity and any ancillary information. AccessMonitor determines whether any given access is permitted,
    given the AuthToken and a per-object auth.Guard, which encapsulates an access control list.
    """
    def __init__(self):
        return

    def check_reserve(self, *, guard: Guard, requester: AuthToken):
        guard.check_reserve(requester=requester)

    def check_update(self, *, guard: Guard, requester: AuthToken):
        guard.check_update(requester=requester)

    def check_proxy(self, *, proxy: AuthToken, requester: AuthToken):
        if requester is None:
            return proxy

        # TODO
        return requester
